﻿using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Core.Views;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.MainForm;
using HeuristicLab.Visualization;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Views {
  [View("FragmentGraphView")]
  [Content(typeof(IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>>), IsDefaultView = true)]
  public sealed partial class FragmentGraphView : ItemView {
    private int PreferredSpacing = 50;

    private ReingoldTilfordLayoutEngine<TileLayoutNode> layoutEngine;
    private ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode> symbolicExpressionEngine;

    private Dictionary<IGenealogyGraphNode<IFragment<ISymbolicExpressionTreeNode>>, TileLayoutNode> tileDictionary;

    private SymbolicExpressionTreeTile Root { get; set; }

    public new IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>> Content {
      get { return (IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>>)base.Content; }
      set { base.Content = value; }
    }

    public FragmentGraphView() {
      InitializeComponent();

      layoutEngine = new ReingoldTilfordLayoutEngine<TileLayoutNode>(n => n.Children) {
        HorizontalSpacing = 50,
        VerticalSpacing = 50,
      };
      symbolicExpressionEngine = new ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode>(n => n.Subtrees) {
        HorizontalSpacing = 50,
        VerticalSpacing = 50,
        NodeWidth = 80,
        NodeHeight = 40
      };
      tileDictionary = new Dictionary<IGenealogyGraphNode<IFragment<ISymbolicExpressionTreeNode>>, TileLayoutNode>();
    }

    private void MakeTiles() {
      var chart = symbolicExpressionChartControl.Chart;
      tileDictionary.Clear();
      foreach (var node in Content.Nodes) {
        var tile = new SymbolicExpressionTreeTile(chart);
        tile.LayoutEngine = symbolicExpressionEngine;
        tile.Root = node.Content.Root;
        var tileNode = new TileLayoutNode { Tile = tile };
        tileDictionary.Add(node, tileNode);
      }
      foreach (var node in Content.Nodes.Where(n => n.Children.Any())) {
        var layoutNode = tileDictionary[node];
        layoutNode.Children = new List<TileLayoutNode>(node.Children.Select(x => tileDictionary[x]));
      }
    }

    private void Draw() {
      var chart = symbolicExpressionChartControl.Chart;
      var nodes = Content.Nodes.ToList();
      var root = nodes[0];
      var fragmentRoot = tileDictionary[root];
      int maxTileWidth = 0, maxTileHeight = 0;
      var tiles = nodes.Select(x => tileDictionary[x].Tile).ToList();

      foreach (var tile in tiles) {
        var size = tile.Size;
        if (maxTileWidth < size.Width) maxTileWidth = size.Width;
        if (maxTileHeight < size.Height) maxTileHeight = size.Height;
      }
      layoutEngine.NodeWidth = maxTileWidth;
      layoutEngine.NodeHeight = maxTileHeight;
      layoutEngine.HorizontalSpacing = PreferredSpacing;
      layoutEngine.VerticalSpacing = PreferredSpacing;

      var visualNodes = layoutEngine.CalculateLayout(fragmentRoot);

      symbolicExpressionChartControl.UpdateEnabled = false;
      foreach (var visualNode in visualNodes) {
        var tile = visualNode.Content.Tile;
        tile.Position = new Point(visualNode.X, visualNode.Y);
        symbolicExpressionChartControl.Add(tile);
      }

      // add connections between the tiles
      foreach (var node in nodes) {
        var aTile = tileDictionary[node].Tile;
        var aSize = aTile.Size;
        var aPos = aTile.Position;

        var lowerLeftMarker = new FixedSizeCircle(chart, aTile.Position, 5, new Pen(Color.Red), new SolidBrush(Color.Transparent));
        var upperRightMarker = new FixedSizeCircle(chart, new PointD(aPos.X + aSize.Width, aPos.Y + aSize.Height), 5,
          new Pen(Color.Blue), new SolidBrush(Color.Transparent));
        symbolicExpressionChartControl.Add(lowerLeftMarker); symbolicExpressionChartControl.Add(upperRightMarker);

        foreach (var child in node.Children) {
          var bTile = tileDictionary[child].Tile;
          var bSize = bTile.Size;
          var bPos = bTile.Position;

          var line = new Line(chart, new PointD(aPos.X + aSize.Width / 2.0, aPos.Y + aSize.Height), new PointD(bPos.X + bSize.Width / 2.0, bPos.Y));
          symbolicExpressionChartControl.Add(line);
        }
      }
      //      symbolicExpressionChartControl.FlipVertical();
      symbolicExpressionChartControl.UpdateEnabled = true;
      symbolicExpressionChartControl.EnforceUpdate();
    }

    protected override void DeregisterContentEvents() {
      // TODO: Deregister your event handlers here
      base.DeregisterContentEvents();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      // TODO: Register your event handlers here
    }

    #region Event Handlers (Content)
    // TODO: Put event handlers of the content here
    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content != null) {
        MakeTiles();
        Draw();
      }
    }
    #endregion

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      // TODO: Enable or disable controls based on whether the content is null or the view is set readonly
    }

    #region Event Handlers (child controls)

    // TODO: Put event handlers of child controls here.

    #endregion
  }

  internal class TileLayoutNode {
    public SymbolicExpressionTreeTile Tile { get; set; }

    private List<TileLayoutNode> children;
    public IEnumerable<TileLayoutNode> Children {
      get { return children ?? Enumerable.Empty<TileLayoutNode>(); }
      set { children = value.ToList(); }
    }
  }
}